from dsmr_parser.parsers import TelegramParser
from dsmr_parser.exceptions import ParseError, InvalidChecksumError
from dsmr_parser.clients.telegram_buffer import TelegramBuffer
from dsmr_parser import telegram_specifications
from dsmr_parser.clients.settings import (
    SERIAL_SETTINGS_V2_2,
    SERIAL_SETTINGS_V4,
    SERIAL_SETTINGS_V5,
)

import serial

import threading
from omnik.ha_logger import hybridlogger
import time
import socket


class Terminal(object):
    def __init__(
        self,
        config,
        logger,
        hass_api,
        terminal_name,
        dsmr_serial_callback,
        dsmr_version,
    ):
        self.config = config
        self.logger = logger
        self.hass_api = hass_api
        self.terminal_name = terminal_name
        self.dsmr_serial_callback = dsmr_serial_callback
        self.mode = self.config.get(f"dsmr.{self.terminal_name}", "mode", "device")
        if self.mode not in ["tcp", "device"]:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"DSMR terminal {self.terminal_name} mode {self.mode} is not valid. "
                "Should be tcp or device. Ignoring DSMR configuration!",
            )
            return
        self.device = self.config.get(
            f"dsmr.{self.terminal_name}", "device", "/dev/ttyUSB0"
        )
        self.host = self.config.get(f"dsmr.{self.terminal_name}", "host", "localhost")
        self.port = self.config.get(f"dsmr.{self.terminal_name}", "port", "3333")
        self.dsmr_version = dsmr_version
        # start terminal
        self.stop = False
        hybridlogger.ha_log(
            self.logger,
            self.hass_api,
            "INFO",
            f"Initializing DSMR termimal '{terminal_name}'. Mode: {self.mode}.",
        )
        if self.mode == "tcp":
            self.thr = threading.Thread(
                target=self._run_tcp_terminal, name=self.terminal_name
            )
        elif self.mode == "device":
            self.thr = threading.Thread(
                target=self._run_serial_terminal, name=self.terminal_name
            )
        self.thr.start()

    def terminate(self):
        self.stop = True
        # Wait for Thread to shutdown
        self.thr.join()

    def _get_dsmr_parser(self):
        dsmr_version = self.dsmr_version
        if dsmr_version == "2.2":
            specification = telegram_specifications.V2_2
            serial_settings = SERIAL_SETTINGS_V2_2
        elif dsmr_version == "4":
            specification = telegram_specifications.V4
            serial_settings = SERIAL_SETTINGS_V4
        elif dsmr_version == "5":
            specification = telegram_specifications.V5
            serial_settings = SERIAL_SETTINGS_V5
        elif dsmr_version == "5B":
            specification = telegram_specifications.BELGIUM_FLUVIUS
            serial_settings = SERIAL_SETTINGS_V5
        else:
            raise NotImplementedError(
                "No telegram parser found for version: %s", dsmr_version
            )
        self.telegram_parser = TelegramParser(specification)
        serial_settings["timeout"] = 10
        self.serial_settings = serial_settings

    def _dsmr_data_received(self, data):
        """Add incoming data to buffer."""
        data = data.decode("ascii")
        self.telegram_buffer.append(data)

        for telegram in self.telegram_buffer.get_all():
            self._handle_telegram(telegram)

    def _handle_telegram(self, telegram):
        """Send off parsed telegram to handling callback."""
        try:
            parsed_telegram = self.telegram_parser.parse(telegram)
        except InvalidChecksumError as e:
            self.logger.warning(str(e))
        except ParseError:
            self.logger.exception("failed to parse telegram")
        else:
            self.dsmr_serial_callback(parsed_telegram)

    def _run_tcp_terminal(self):
        self._get_dsmr_parser()
        # buffer to keep incomplete incoming data
        while not self.stop:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"DSMR terminal {self.terminal_name} was started.",
            )
            try:
                self.telegram_buffer = TelegramBuffer()
                server_address = (self.host, self.port)
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.sock.settimeout(5)
                self.sock.connect(server_address)
                while not self.stop:
                    data = self.sock.recv(1024)
                    if not data and not self.stop:
                        hybridlogger.ha_log(
                            self.logger,
                            self.hass_api,
                            "INFO",
                            f"DSMR terminal {self.terminal_name} was interrupted and will be restarted.",
                        )
                        time.sleep(5)
                        break
                    elif data:
                        self._dsmr_data_received(data)

            except Exception as e:
                if not self.stop:
                    hybridlogger.ha_log(
                        self.logger,
                        self.hass_api,
                        "WARNING",
                        f"DSMR terminal {self.terminal_name} was interrupted "
                        f"and will be restarted in a few moments: {e.args}",
                    )
                    time.sleep(5)
            finally:
                self.sock.close()

    def _run_serial_terminal(self):
        self._get_dsmr_parser()
        # buffer to keep incomplete incoming data
        while not self.stop:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"DSMR terminal {self.terminal_name} was started.",
            )
            try:
                self.telegram_buffer = TelegramBuffer()
                self.sock = serial.Serial(port=self.device, **self.serial_settings)
                while not self.stop:
                    data = self.sock.read_until()
                    if not data and not self.stop:
                        hybridlogger.ha_log(
                            self.logger,
                            self.hass_api,
                            "INFO",
                            f"DSMR terminal {self.terminal_name} was interrupted and will be restarted.",
                        )
                        time.sleep(5)
                        break
                    elif data:
                        self._dsmr_data_received(data)

            except Exception as e:
                if not self.stop:
                    hybridlogger.ha_log(
                        self.logger,
                        self.hass_api,
                        "WARNING",
                        f"DSMR terminal {self.terminal_name} was interrupted "
                        f"and will be restarted in a few moments: {e.args}",
                    )
                    time.sleep(5)
            finally:
                self.sock.close()



import os
import sys
import logging

from requests.exceptions import RequestException
from omnik.ha_logger import hybridlogger
from omnik.plant import Plant
import requests
import json
from .daylight import daylight
import threading
import time
from decimal import Decimal
from datetime import date, datetime, timedelta, timezone

from .plugin_output import Plugin
from .plugin_client import Client
from .dsmr import DSRM
from copy import deepcopy
from cachetools import Cache

import importlib

# from .plugin_localproxy import LocalProxyPlugin

logger = logging.getLogger(__name__)


class DataLogger(object):
    def __init__(self, config, hass_api=None):
        # Defaults to UTC now() - every interval
        self.plant_update = {}
        self.hass_api = hass_api
        self.logger = logger
        self.config = config
        self.cache = Cache(maxsize=10)
        self.start_total_energy = {}
        self.pasttime = datetime.combine(date.today(), datetime.min.time()).timestamp()
        self.dsmr_access = threading.Condition(threading.Lock())

        if self.config.get("default", "debug", fallback=False):
            logger.setLevel(logging.DEBUG)

        if not self._init_data_fields():
            sys.exit(1)
        # init energy cache
        # total_energy_cache.json
        self.persistant_cache_file = self.config.get(
            "default", "persistant_cache_file", fallback="./persistant_cache.json"
        )
        self._load_persistant_cache()
        # read data_fields
        try:
            with open(self.data_config_file) as json_file_config:
                self.config.data_field_config = json.load(json_file_config)
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"Error reading configuration file (data_fields.json). Exiting!. Error: {e.args}",
            )
            raise e

        # read attributes
        if not self._read_attributes():
            sys.exit(1)

        self.every = self._get_interval()
        self.interval_aggregated = self._get_interval_aggregated()

        # Make sure we check for a recent update first
        self.last_update_time = datetime.now(timezone.utc) - timedelta(
            seconds=self.interval_aggregated
        )

        # Initialize plugin paths
        sys.path.append(self.__expand_path("plugin_output"))
        sys.path.append(self.__expand_path("plugin_client"))
        sys.path.append(self.__expand_path("plugin_localproxy"))

        self.city = self.config.get("default", "city", fallback="Amsterdam")
        try:
            self.dl = daylight(self.city)
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"City '{self.city}' not recognized. Error: {e}",
            )
            sys.exit(1)

        # Initialize client
        self._init_client()
        # Initialize output plugins
        self._init_output_plugins()

        self.omnik_api_level = 0

        # Init dsmr
        self._init_dsmr()

    def _init_data_fields(self):
        self.config.data_config_file_args = self.config.get(
            "default", "data_config", fallback=""
        )
        hybridlogger.ha_log(
            self.logger,
            self.hass_api,
            "DEBUG",
            f"Data configuration [args]: '{self.config.data_config_file_args}'.",
        )
        hybridlogger.ha_log(
            self.logger,
            self.hass_api,
            "DEBUG",
            f"Data configuration [path]: '{self.config.data_config_file_path}'.",
        )
        hybridlogger.ha_log(
            self.logger,
            self.hass_api,
            "DEBUG",
            f"Data configuration [shared]: '{self.config.data_config_file_shared}'.",
        )
        hybridlogger.ha_log(
            self.logger,
            self.hass_api,
            "DEBUG",
            f"Data configuration [cwd]: '{self.config.data_config_file_cwd}'.",
        )
        if self.config.configfile:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Using configuration from '{self.config.configfile}'.",
            )

        if os.path.exists(self.config.data_config_file_args):
            self.data_config_file = self.config.data_config_file_args
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Using configured data configuration from '{self.data_config_file}'.",
            )
        elif os.path.exists(self.config.data_config_file_path):
            self.data_config_file = self.config.data_config_file_path
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Using configured data configuration from '{self.data_config_file}'.",
            )
        elif os.path.exists(self.config.data_config_file_shared):
            self.data_config_file = self.config.data_config_file_shared
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Using shared data configuration from '{self.data_config_file}'.",
            )
        elif os.path.exists(self.config.data_config_file_cwd):
            self.data_config_file = self.config.data_config_file_cwd
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Using shared data configuration current working directory '{self.data_config_file}'.",
            )
        else:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                "No valid data configuration file (data_fields.json) found. Exiting!",
            )
            return False
        return True

    def _init_attribute_dict(self):
        self.config.attributes = {}
        asset_classes = list(["omnik", "omnik_dsmr", "dsmr", "dsmr_gas"])
        self.config.attributes["asset"] = {
            "omnik": list(["inverter", "plant_id", "last_update"]),
            "omnik_dsmr": list(
                ["inverter", "plant_id", "EQUIPMENT_IDENTIFIER", "last_update"]
            ),
            "dsmr": list(["EQUIPMENT_IDENTIFIER", "timestamp"]),
            "dsmr_gas": list(["EQUIPMENT_IDENTIFIER_GAS", "timestamp_gas"]),
        }
        self.config.attributes["model"] = {
            "omnik": "Omniksol",
            "omnik_dsmr": "Omnik data logger",
            "dsmr": "DSMR electicity meter",
            "dsmr_gas": "DSMR gas meter",
        }
        self.config.attributes["mf"] = {
            "omnik": "Omnik",
            "omnik_dsmr": "JBsoft",
            "dsmr": "Netbeheer Nederland",
            "dsmr_gas": "Netbeheer Nederland",
        }
        self.config.attributes["identifier"] = {
            "omnik": "plant_id",
            "omnik_dsmr": "plant_id",
            "dsmr": "EQUIPMENT_IDENTIFIER",
            "dsmr_gas": "EQUIPMENT_IDENTIFIER_GAS",
        }
        self.config.attributes["devicename"] = {
            "omnik": "Inverter",
            "omnik_dsmr": "Omnik_data_logger",
            "dsmr": "DSMR_electicity_meter",
            "dsmr_gas": "DSMR_gasmeter",
        }
        return asset_classes

    def _init_attribute_set(self, attribute, asset_class, type="string"):
        if attribute not in self.config.attributes:
            self.config.attributes[attribute] = {}
        if not self.config.has_option("attributes", f"{attribute}.{asset_class}"):
            return
        if type == "list":
            self.config.attributes[attribute][asset_class] = self.config.getlist(
                "attributes", f"asset.{asset_class}"
            )
        else:
            self.config.attributes[attribute][asset_class] = self.config.get(
                "attributes", f"{attribute}.{asset_class}"
            )

    def _read_attributes(self):
        asset_classes = self._init_attribute_dict()
        try:
            if self.config.has_option("attributes", "asset_classes"):
                asset_classes += list(
                    set(self.config.getlist("attributes", "asset_classes"))
                    - set(asset_classes)
                )
            for asset_class in asset_classes:
                self._init_attribute_set("asset", asset_class, "list")
                self._init_attribute_set("model", asset_class)
                self._init_attribute_set("mf", asset_class)
                self._init_attribute_set("identifier", asset_class)
                self._init_attribute_set("devicename", asset_class)
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"Error reading attributes from config. Error: {e.args}",
            )
            raise e
        return True

    def terminate(self):
        # Cleanup Client
        self._terminate_client()
        # Cleanup Output plugins
        self._terminate_output_plugins()
        # Terminate DSMR
        if self.dsmr:
            self.dsmr.terminate()

    def _init_client(self):
        # For now the default client is not set by default, client should be configured in the config
        # options are: localproxy, solarmanpv and omnikportal
        self.client_module = self.config.get("plugins", "client", None)
        if not self.client_module:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                "No 'client' parameter configured under the [plugin] section.",
            )
            sys.exit(1)

        hybridlogger.ha_log(
            self.logger,
            self.hass_api,
            "INFO",
            f"Initializing client : {self.client_module}.",
        )
        # Try to load client
        try:
            Client.logger = self.logger
            Client.config = self.config
            Client.hass_api = self.hass_api
            # __import__(self.client_module)
            spec = importlib.util.find_spec(self.client_module)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            self.client = Client.client[0]
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"Client '{self.client_module}' could not be initialized. Error: {e}",
            )
            sys.exit(1)

    def _terminate_client(self):
        while Client.client:
            Client.client.pop(0).terminate()

    def _init_output_plugins(self):
        self.plugins = self.config.getlist("plugins", "output", fallback=[""])
        if self.plugins and self.plugins[0]:
            hybridlogger.ha_log(
                self.logger, self.hass_api, "INFO", f"Plugins enabled: {self.plugins}."
            )
        else:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                "No output plugins configured! Monitoring only.",
            )
        # Import output plugins
        if self.plugins and self.plugins[0]:
            Plugin.logger = self.logger
            Plugin.config = self.config
            Plugin.hass_api = self.hass_api

            for plugin in self.plugins:
                spec = importlib.util.find_spec(plugin)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                # importlib.import_module(plugin)

    def _terminate_output_plugins(self):
        while Plugin.plugins:
            Plugin.plugins.pop(0).terminate()

    def _init_dsmr(self):
        terminals = self.config.getlist("dsmr", "terminals", fallback=[])
        if terminals and terminals[0]:
            self.dsmr = DSRM()
            self.dsmr.config = self.config
            self.dsmr.logger = self.logger
            self.dsmr.hass_api = self.hass_api
            self.dsmr.datalogger = self
            self.dsmr.initialize(terminals=terminals, dsmr_callback=self.dsmr_callback)
        else:
            self.dsmr = None
        self.dsmr_data = {}

    def _get_dsmr_data(self, plant, data):
        # if dsmr measurements are not enabled then return
        if not self.dsmr:
            return
        # Try to merge with the latest dsmr data available
        complete = False
        tries = 20
        if "plant_id" not in data:
            data["plant_id"] = "0"
        if plant not in self.dsmr_data:
            return False
        while tries > 0:
            self.dsmr_access.acquire()
            if plant in self.dsmr_data:
                # Insert raw DSMR data
                data.update(self._dsmr_cache(plant, data["last_update"]))
                # Insert calculated netto values (solar - net)
                self._calculate_consumption(data)
                data["total_energy"] = data.pop("total_energy_recalc")
                complete = True
            self.dsmr_access.release()
            tries -= 1
            if not complete:
                time.sleep(1)
            else:
                break
        if not tries:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"Could not match DSMR data. "
                "Did you configure your dsmr terminal correctly?",
            )
        return complete

    def _calculate_consumption(self, data):
        # Calculate cumulative energy direct used
        data["energy_direct_use"] = (
            data["total_energy_recalc"]
            - data["total_energy_offset"]
            - data["energy_delivered_net"]
        )
        data["energy_used"] = (
            data["energy_used_net"] + data["energy_direct_use"]
        )  # use as v3 * 1000 for pvoutput
        # Calculate the actual power used
        if "current_power" in data:
            data["power_direct_use"] = data["current_power"] - (
                data["CURRENT_ELECTRICITY_DELIVERY"] * Decimal("1000")
            )
            # Should not be negative, but timestamps can be out of sync.
            if data["power_direct_use"] < Decimal("0"):
                data["power_direct_use"] = Decimal("0")
            data["power_consumption"] = (
                data["CURRENT_ELECTRICITY_USAGE"] * Decimal("1000")
                + data["power_direct_use"]
            )
        else:
            data["power_consumption"] = data["CURRENT_ELECTRICITY_USAGE"] * Decimal(
                "1000"
            )
        data["last_update_calc"] = time.time()

    def dsmr_callback(self, terminal, dsmr_message):
        # print(f"{dsmr_message['gas_consumption_totall']} =
        # m3 {dsmr_message['gas_consumption_hour']} m3/h - {dsmr_message['timestamp_gas']}")
        if dsmr_message["plant_id"]:
            plant_id = dsmr_message.pop("plant_id")
        else:
            plant_id = "0"
        # Get last total energy form cache (if available) and ad to the dataset
        # dsmr_message['total_energy_cache'] = self.total_energy(plant_id)
        self.dsmr_access.acquire()
        self._dsmr_cache_update(plant_id, dsmr_message)
        # self.dsmr_data[plant_id] = deepcopy(dsmr_message)
        self.dsmr_access.release()
        if plant_id in self.plant_update:
            # DSMR for specific plant
            self._proces_pushed_net_event(plant_id, dsmr_message)
        else:
            # Aggregated DSMR only
            self._proces_pushed_net_event("0", dsmr_message)

    def _dsmr_cache_update(self, plant_id, dsmr_message):
        if plant_id not in self.dsmr_data:
            self.dsmr_data[plant_id] = list()
        if len(self.dsmr_data[plant_id]) == 10:
            self.dsmr_data[plant_id].pop(0)
        self.dsmr_data[plant_id].append(dsmr_message)

    def _dsmr_cache(self, plant_id, timestamp):
        max_delta = -1
        best_data = None
        for data in reversed(self.dsmr_data[plant_id]):
            dsmr_ts = data["timestamp"]
            delta = abs(timestamp - dsmr_ts)
            if delta < max_delta or max_delta < 0:
                max_delta = delta
                best_data = data
        return best_data

    def _get_interval(self):
        if self.config.has_option("default", "interval"):
            return int(self.config.get("default", "interval"))
        else:
            return 0

    def _get_interval_aggregated(self):
        if self.config.has_option("default", "interval_aggregated"):
            return int(self.config.get("default", "interval_aggregated"))
        else:
            return 300

    def _validate_user_login(self):
        self.omnik_api_level = 0
        try:
            self.client.initialize()
            # Logged on
            self.omnik_api_level = 1
        except RequestException as errc:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"Connection error during account validation omnik portal: {errc}",
            )
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", e)

    def _logon(self):
        # Log on to the omnik portal to enable fetching plant's
        if self.omnik_api_level == 0:
            self._validate_user_login()
            if self.omnik_api_level == 0:
                return False
        return True

    def _fetch_plants(self):
        # Fetch te plant's available for the account
        if not self.plant_update or self.omnik_api_level == 1:
            try:
                plants = self.client.getPlants()
                for pid in plants:
                    self.plant_update[str(pid["plant_id"])] = Plant(
                        pid["plant_id"], self.last_update_time.astimezone(timezone.utc)
                    )
                self.omnik_api_level = 2
            except requests.exceptions.RequestException as err:
                hybridlogger.ha_log(
                    self.logger, self.hass_api, "WARNING", f"Request error: {err}"
                )
                self.omnik_api_level = 0
                return False
            except Exception as e:
                hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", e)
                self.omnik_api_level = 0
                return False
        return True

    def _listen_for_update(self):
        data = self.client.getPlantData()
        return data

    def _process_received_update(self, data, netdata=False, plant=None):
        # Get the report time
        newreporttime = datetime.fromtimestamp(data["last_update"]).astimezone(
            timezone.utc
        )
        if not plant:
            plant = data["plant_id"]
        if plant == "0":
            updated_entities = "aggregated logging"
        else:
            updated_entities = f"plant {plant}"
        if netdata:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Net data update for {updated_entities} UTC {newreporttime}",
            )
        else:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Update for {updated_entities} UTC {newreporttime}",
            )

        return plant

    def _sundown_reset_power(self, data):
        if self.client.use_timer:
            if self.sundown:
                data["current_power"] = Decimal("0")

    def _fetch_update(self, plant):
        try:
            data = self.client.getPlantData(plant)
            if data:
                self._sundown_reset_power(data)
                # Get the actual report time from the omnik portal
                newreporttime = datetime.fromtimestamp(data["last_update"]).astimezone(
                    timezone.utc
                )
                # Only proces updates that occured after we started or start a single measurement (TODO)
                if (
                    newreporttime > self.plant_update[plant].last_update_time
                    or not self.every
                    or self.sundown
                ):
                    hybridlogger.ha_log(
                        self.logger,
                        self.hass_api,
                        "INFO",
                        f"Update for plant {plant} UTC {newreporttime}",
                    )
                    # the newest plant_update time will be used as a baseline to program the timer
                    self.last_update_time = newreporttime
                    # Store the plant_update time for each indiviual plant,
                    # update only if there is a new report available to avoid duplicates
                    self.plant_update[
                        plant
                    ].last_update_time = newreporttime.astimezone(timezone.utc)
                    data["plant_id"] = plant
                    return data
                else:
                    hybridlogger.ha_log(
                        self.logger,
                        self.hass_api,
                        "INFO",
                        f"No recent report update to process. Last report at UTC {newreporttime}",
                    )
                    return None

        except requests.exceptions.RequestException as err:
            hybridlogger.ha_log(
                self.logger, self.hass_api, "WARNING", f"Request error: {err}"
            )
            self.omnik_api_level = 0
            # Abort retry later
            return None
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"Unexpected error during data handling: {e}",
            )
            self.omnik_api_level = 0
            # Abort retry later
            return None

    def _validate_field(self, data, field, fallback):
        # validate existance of field in dict
        # provide a default value if it is not available
        if data:
            if field not in data:
                data[field] = fallback

    def _validate_client_data(self, plant, data):
        # Insert defaults for missing fields
        # data['data']['plant_id']
        self._validate_field(
            data, "inverter", self.config.get(f"plant.{plant}", "inverter_sn", "n/a")
        )
        return data

    def _output_update(self, plant, data):
        # Insert dummy data for fields that have not been supplied by the client
        data = self._validate_client_data(plant, data)
        # Process for each plugin, but only when valid
        for plugin in Plugin.plugins:
            if not plugin.process_aggregates or "sys_id" in data:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "DEBUG",
                    f"Trigger plugin '{getattr(plugin, 'name')}'.",
                )
                plugin.process(msg=data)

    def _output_update_aggregated_data(self, plant, data):
        # Insert dummy data for fields that have not been supplied by the client
        data = self._validate_client_data(plant, data)
        # Process for each plugin, but only when valid
        for plugin in Plugin.plugins:
            if plugin.process_aggregates:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "DEBUG",
                    f"Trigger plugin '{getattr(plugin, 'name')}' with aggregated data.",
                )
                plugin.process(msg=data)

    def _init_aggregated_data(self, aggregated_data, data, sys_id):
        if not aggregated_data:
            # Initialize
            aggregated_data["sys_id"] = sys_id
            for field in [
                "last_update",
                "current_power",
                "today_energy",
                "total_energy",
                "voltage_ac1",
                "voltage_ac2",
                "voltage_ac3",
                "voltage_ac_max",
                "inverter_temperature",
            ]:
                self._init_aggregated_data_field(aggregated_data, data, field)
            if self.dsmr:
                for field in [
                    # Attributes Electricity for pvoutput
                    "power_consumption",
                    "energy_used",
                    "energy_used_net",
                    "net_voltage_max",
                    "INSTANTANEOUS_VOLTAGE_L1",
                    "INSTANTANEOUS_VOLTAGE_L2",
                    "INSTANTANEOUS_VOLTAGE_L3",
                    # Other attributes Electricity for mqtt or influxdb
                    "timestamp",
                    "ELECTRICITY_USED_TARIFF_1",
                    "ELECTRICITY_USED_TARIFF_2",
                    "ELECTRICITY_DELIVERED_TARIFF_1",
                    "ELECTRICITY_DELIVERED_TARIFF_2",
                    "energy_delivered_net",
                    "CURRENT_ELECTRICITY_USAGE",
                    "CURRENT_ELECTRICITY_DELIVERY",
                    "ELECTRICITY_ACTIVE_TARIFF",
                    "LONG_POWER_FAILURE_COUNT",
                    "SHORT_POWER_FAILURE_COUNT",
                    "VOLTAGE_SAG_L1_COUNT",
                    "VOLTAGE_SAG_L2_COUNT",
                    "VOLTAGE_SAG_L3_COUNT",
                    "VOLTAGE_SWELL_L1_COUNT",
                    "VOLTAGE_SWELL_L2_COUNT",
                    "VOLTAGE_SWELL_L3_COUNT",
                    "INSTANTANEOUS_ACTIVE_POWER_L1_POSITIVE",
                    "INSTANTANEOUS_ACTIVE_POWER_L1_NEGATIVE",
                    "INSTANTANEOUS_ACTIVE_POWER_L2_POSITIVE",
                    "INSTANTANEOUS_ACTIVE_POWER_L2_NEGATIVE",
                    "INSTANTANEOUS_ACTIVE_POWER_L3_POSITIVE",
                    "INSTANTANEOUS_ACTIVE_POWER_L3_NEGATIVE",
                    "current_net_power",
                    "current_net_power_l1",
                    "current_net_power_l2",
                    "current_net_power_l3",
                    "net_voltage_max",
                    "INSTANTANEOUS_CURRENT_L1",
                    "INSTANTANEOUS_CURRENT_L2",
                    "INSTANTANEOUS_CURRENT_L3",
                    "net_current_l1",
                    "net_current_l2",
                    "net_current_l3",
                    # Other attributes Gas
                    "timestamp_gas",
                    "gas_consumption_total",
                    "gas_consumption_hour",
                ]:
                    self._init_aggregated_data_field(aggregated_data, data, field)
                # Other attributes Electricity
                # Other attributes Gas
                {
                    "timestamp_gas": 1622714706.0,
                    "gas_consumption_total": 2864.6,
                    "gas_consumption_hour": 0,
                }

    def _init_aggregated_data_field(self, aggregated_data, data, field):
        if field in data:
            aggregated_data[field] = Decimal("0")

    def _adapt_max_value(self, aggregated_data, data, field):
        if field in data:
            if data[field] > aggregated_data[field]:
                aggregated_data[field] = data[field]

    def _adapt_add_value(self, aggregated_data, data, field):
        if field in data:
            aggregated_data[field] += data[field]

    def _adapt_last_value(self, aggregated_data, data, field):
        if field in data:
            aggregated_data[field] = data[field]

    def _aggregate_data(self, aggregated_data, data):
        if not data:
            return
        # Check for pvoutput sys_id override in config
        sys_id = int(self.config.get(f"plant.{data['plant_id']}", "sys_id", "0"))
        global_sys_id = int(self.config.get("output.pvoutput", "sys_id", "0"))
        if sys_id and (sys_id != global_sys_id):
            # Do no aggegate: add sys_id to data set
            data["sys_id"] = sys_id
        else:
            # Get sys_id from pvoutput section, cannot aggregate without sys_id
            sys_id = global_sys_id
            if sys_id:
                # Aggerate data (initialize dict and set sys_id)
                self._init_aggregated_data(aggregated_data, data, sys_id)
                # Timestamp
                self._adapt_max_value(aggregated_data, data, "last_update")
                # Add energy
                self._adapt_add_value(aggregated_data, data, "today_energy")
                self._adapt_add_value(aggregated_data, data, "total_energy")
                # Add power
                self._adapt_add_value(aggregated_data, data, "current_power")
                # Max voltage
                self._adapt_max_value(aggregated_data, data, "voltage_ac_max")
                self._adapt_max_value(aggregated_data, data, "voltage_ac1")
                self._adapt_max_value(aggregated_data, data, "voltage_ac2")
                self._adapt_max_value(aggregated_data, data, "voltage_ac3")
                self._adapt_max_value(aggregated_data, data, "net_voltage_max")
                # Max inverter temperature
                self._adapt_max_value(aggregated_data, data, "inverter_temperature")
            if sys_id and self.dsmr:
                self._adapt_add_value(aggregated_data, data, "power_consumption")
                self._adapt_add_value(aggregated_data, data, "energy_used_net")
                self._adapt_add_value(aggregated_data, data, "energy_used")
                # DSMR specific data for output to mqtt or influx with multiple meters or inverters
                self._adapt_max_value(aggregated_data, data, "INSTANTANEOUS_VOLTAGE_L1")
                self._adapt_max_value(aggregated_data, data, "INSTANTANEOUS_VOLTAGE_L2")
                self._adapt_max_value(aggregated_data, data, "INSTANTANEOUS_VOLTAGE_L3")
                # Other attributes Electricity for mqtt or influxdb
                self._adapt_last_value(aggregated_data, data, "timestamp")
                self._adapt_last_value(
                    aggregated_data, data, "ELECTRICITY_USED_TARIFF_1"
                )
                self._adapt_last_value(
                    aggregated_data, data, "ELECTRICITY_USED_TARIFF_2"
                )
                self._adapt_last_value(
                    aggregated_data, data, "ELECTRICITY_DELIVERED_TARIFF_1"
                )
                self._adapt_last_value(
                    aggregated_data, data, "ELECTRICITY_DELIVERED_TARIFF_2"
                )
                self._adapt_last_value(
                    aggregated_data, data, "ELECTRICITY_ACTIVE_TARIFF"
                )

                self._adapt_add_value(aggregated_data, data, "energy_delivered_net")
                self._adapt_add_value(
                    aggregated_data, data, "CURRENT_ELECTRICITY_USAGE"
                )
                self._adapt_add_value(
                    aggregated_data, data, "CURRENT_ELECTRICITY_DELIVERY"
                )
                self._adapt_add_value(aggregated_data, data, "LONG_POWER_FAILURE_COUNT")
                self._adapt_add_value(
                    aggregated_data, data, "SHORT_POWER_FAILURE_COUNT"
                )
                self._adapt_add_value(aggregated_data, data, "VOLTAGE_SAG_L1_COUNT")
                self._adapt_add_value(aggregated_data, data, "VOLTAGE_SAG_L2_COUNT")
                self._adapt_add_value(aggregated_data, data, "VOLTAGE_SAG_L3_COUNT")
                self._adapt_add_value(aggregated_data, data, "VOLTAGE_SWELL_L1_COUNT")
                self._adapt_add_value(aggregated_data, data, "VOLTAGE_SWELL_L2_COUNT")
                self._adapt_add_value(aggregated_data, data, "VOLTAGE_SWELL_L3_COUNT")

                self._adapt_add_value(
                    aggregated_data, data, "INSTANTANEOUS_ACTIVE_POWER_L1_POSITIVE"
                )
                self._adapt_add_value(
                    aggregated_data, data, "INSTANTANEOUS_ACTIVE_POWER_L2_POSITIVE"
                )
                self._adapt_add_value(
                    aggregated_data, data, "INSTANTANEOUS_ACTIVE_POWER_L3_POSITIVE"
                )
                self._adapt_add_value(
                    aggregated_data, data, "INSTANTANEOUS_ACTIVE_POWER_L1_NEGATIVE"
                )
                self._adapt_add_value(
                    aggregated_data, data, "INSTANTANEOUS_ACTIVE_POWER_L2_NEGATIVE"
                )
                self._adapt_add_value(
                    aggregated_data, data, "INSTANTANEOUS_ACTIVE_POWER_L3_NEGATIVE"
                )

                self._adapt_add_value(aggregated_data, data, "current_net_power")
                self._adapt_add_value(aggregated_data, data, "current_net_power_l1")
                self._adapt_add_value(aggregated_data, data, "current_net_power_l2")
                self._adapt_add_value(aggregated_data, data, "current_net_power_l3")
                self._adapt_add_value(aggregated_data, data, "INSTANTANEOUS_CURRENT_L1")
                self._adapt_add_value(aggregated_data, data, "INSTANTANEOUS_CURRENT_L2")
                self._adapt_add_value(aggregated_data, data, "INSTANTANEOUS_CURRENT_L3")
                self._adapt_add_value(aggregated_data, data, "net_current_l1")
                self._adapt_add_value(aggregated_data, data, "net_current_l2")
                self._adapt_add_value(aggregated_data, data, "net_current_l3")
                # Other attributes Gas
                self._adapt_last_value(aggregated_data, data, "timestamp_gas")

                self._adapt_add_value(aggregated_data, data, "gas_consumption_total")
                self._adapt_add_value(aggregated_data, data, "gas_consumption_hour")

    def _digitize(self, data):
        digitize_fields = [
            "current_power",
            "today_energy",
            "total_energy",
            "inverter_temperature",
            "current_ac1",
            "current_ac2",
            "current_ac3",
            "voltage_ac1",
            "voltage_ac2",
            "voltage_ac3",
            "voltage_ac_max",
            "frequency_ac1",
            "frequency_ac2",
            "frequency_ac3",
            "power_ac1",
            "power_ac2",
            "power_ac3",
            "current_pv1",
            "current_pv2",
            "current_pv3",
            "voltage_pv1",
            "voltage_pv2",
            "voltage_pv3",
            "power_pv1",
            "power_pv2",
            "power_pv3",
            "current_power_pv",
            "operation_hours",
        ]
        for field in data:
            if field in digitize_fields:
                data[field] = Decimal(f"{data[field]}")
        self._total_energy_recalc(data)

    def _total_energy_recalc(self, data):
        # Re calculate total energy for today accuracy (TODO per plant_id!) update after every measurement
        # store in cache and make persistant
        # Also store the last current power to the cache
        if "plant_id" in data:
            plant = data["plant_id"]
        else:
            plant = "0"
        data["total_energy_recalc"] = self.total_energy(
            plant,
            today_energy=data["today_energy"],
            total_energy=data["total_energy"],
            current_power=data["current_power"],
        )

    def _load_persistant_cache(self):
        try:
            with open(self.persistant_cache_file) as total_energy_cache:
                file_cache = json.load(total_energy_cache)
                for item in file_cache:
                    self.cache[item] = Decimal(f"{file_cache[item]}")
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Using energy cache file '{self.persistant_cache_file}'.",
            )
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Cache file '{self.persistant_cache_file}' was not used previously. Error: {e.args}",
            )

    def _update_persistant_cache(self):
        try:
            file_cache = {}
            for item in self.cache:
                file_cache[item] = float(self.cache[item])
            with open(self.persistant_cache_file, "w") as json_file_config:
                json.dump(file_cache, json_file_config)
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"Cache file '{self.persistant_cache_file}' can not be written! Error: {e.args}",
            )

    def last_current_power(self, plant):
        last_current_power = f"{plant}.last_current_power"
        if last_current_power not in self.cache:
            return Decimal("0")
        else:
            return self.cache[last_current_power]

    def total_energy(
        self,
        plant,
        today_energy=None,
        total_energy=None,
        current_power=None,
        lifetime=True,
    ):
        # start_total_energy_key = f'{plant}.start_total_energy'
        last_total_energy = f"{plant}.last_total_energy"
        last_today_energy = f"{plant}.last_today_energy"
        last_current_power = f"{plant}.last_current_power"
        # if total energy is supplied, update cache directly
        if total_energy:
            self.cache[last_total_energy] = total_energy
            self.cache[last_today_energy] = today_energy
            self.cache[last_current_power] = current_power
            self._update_persistant_cache()

        if plant not in self.start_total_energy:
            if total_energy:
                self.start_total_energy[plant] = total_energy - today_energy
                return (
                    self.start_total_energy[plant] + today_energy
                    if lifetime
                    else today_energy
                )
            else:
                # No accurate total energy is available, use total_energy cache
                if last_today_energy in self.cache and last_total_energy in self.cache:
                    self.start_total_energy[plant] = (
                        self.cache[last_total_energy] - self.cache[last_today_energy]
                    )
                    return (
                        self.start_total_energy[plant] + self.cache[last_today_energy]
                        if lifetime
                        else self.cache[last_today_energy]
                    )
                else:
                    return None
        else:
            if total_energy and not today_energy:
                self.start_total_energy[plant] = total_energy
                return total_energy
            elif today_energy:
                return self.start_total_energy[plant] + today_energy
            else:
                if last_today_energy in self.cache and last_total_energy in self.cache:
                    return (
                        self.start_total_energy[plant] + self.cache[last_today_energy]
                        if lifetime
                        else self.cache[last_today_energy]
                    )
                else:
                    return None

    def _sunshine_check(self):
        self.sundown = not self.dl.sun_shine()
        if self.sundown:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"No sunshine postponing till down next dawn {self.dl.next_dawn}.",
            )
            # Send 0 Watt update
            return self.dl.next_dawn + timedelta(minutes=10)
        else:
            # return the last report time return value, but not when there is no sun
            return self.last_update_time

    def _proces_pushed_net_event(self, plant_id, dsmr_message):

        # ts_midnight = datetime.combine(date.today(), datetime.min.time()).timestamp()
        data = deepcopy(dsmr_message)
        dsmr_timestamp = data["timestamp"]
        # if (dsmr_timestamp - last_update) > (self.every + 10) and dsmr_timestamp > self.pasttime:
        if dsmr_timestamp > self.pasttime and self.total_energy(plant_id):
            # Process independent net data for aggregated clients with regards of rate limits
            # data['current_power'] = Decimal('0')
            # Omnik related fields need 'last_update' timestamp
            data["last_update"] = dsmr_timestamp
            data["plant_id"] = plant_id
            data["total_energy_recalc"] = self.total_energy(plant_id)
            data["today_energy"] = self.total_energy(plant_id, lifetime=False)
            self._calculate_consumption(data)
            data["total_energy"] = data.pop("total_energy_recalc")

            aggegated_data = {}
            self._aggregate_data(aggegated_data, data)
            # Do not publish the current power, since we do not know the last state
            # set next time block for update
            self.pasttime = (
                dsmr_timestamp
                - (dsmr_timestamp % self.interval_aggregated)
                + self.interval_aggregated
            )
            # proces net updates out for aggegated output (pvoutput) with (multiple) pushing inverters with aggegation and no updates
            last_update = 0.0
            if plant_id == "0":
                for plant in self.plant_update:
                    if (
                        datetime.timestamp(self.plant_update[plant].last_update_time)
                        > last_update
                    ):
                        last_update = datetime.timestamp(
                            self.plant_update[plant].last_update_time
                        )
            else:
                last_update = datetime.timestamp(
                    self.plant_update[plant_id].last_update_time
                )
            # we will send (net) updates when inverers do not retreive new updates
            if (dsmr_timestamp - last_update) > self.every:
                self._process_received_update(
                    aggegated_data, netdata=True, plant=plant_id
                )
                # Export combined to pvoutput
                self._output_update_aggregated_data(plant_id, aggegated_data)
                # Export combined data to other output clients
                self._output_update(plant, aggegated_data)
        # TODO process net update for other clients
        data2 = deepcopy(dsmr_message)
        data2["plant_id"] = plant_id
        self._validate_client_data(plant_id, data2)
        self._output_update(plant_id, data2)

    def _process_timed_event(self):
        # For portal clients and tcpclient with a fixed interval
        # Returns the date time of the last update (time behind) or a time in te future when postposing updates
        # Do a Sunshine check when a Timer is used
        next_report_at = self._sunshine_check()

        # Check for login, if needed. Return None on failure
        if not self._logon():
            # Logon failed
            # Sun is down, skip this measurement and wait till dawn
            return next_report_at if self.sundown else None

        # Caching of plant id's, if needed. Return None on failure
        if not self._fetch_plants():
            # Sun is down, skip this measurement and wait till dawn
            return next_report_at if self.sundown else None

        # Process data reports for each plant
        if self.omnik_api_level == 2:
            aggegated_data = {}
            skip_aggregation = False
            for plant in self.plant_update:
                # Try getting a new update or wait for logging event
                data = self._fetch_update(plant)
                if data:
                    # A new last update time was set
                    if self.plant_update[plant].last_update_time > next_report_at:
                        next_report_at = self.plant_update[plant].last_update_time
                    # Digitize data
                    self._digitize(data)
                    # Get specific dsmr data
                    self._get_dsmr_data(plant, data)
                    # Assemble aggegated data
                    self._aggregate_data(aggegated_data, data)
                    # export the data to the output plugins
                    self._output_update(plant, data)
                else:
                    # Use cached data for aggregation with DSMR
                    data = {}
                    try:
                        data["total_energy"] = self.cache[f"{plant}.last_total_energy"]
                        data["today_energy"] = self.cache[f"{plant}.last_today_energy"]
                        data["current_power"] = Decimal("0.0")
                        data["last_update"] = time.time()
                        # Digitize data
                        self._digitize(data)
                        # Assemble aggegated data
                        self._aggregate_data(aggegated_data, data)
                    except:
                        # do not allow data aggregation unless we have valid data
                        skip_aggregation = True

            # Process aggregated data over all plants if no specific plants are configured with DSMR terminal
            if aggegated_data and not skip_aggregation:
                # Get dsmr data for aggegated data
                self._total_energy_recalc(aggegated_data)
                # Get dsmr data for aggegated data
                self._get_dsmr_data("0", aggegated_data)
                # Output aggregated data to influx/mqtt if we have multiple plants
                if len(self.client.plant_id_list) > 1:
                    self._output_update(plant, aggegated_data)
                self._output_update_aggregated_data(plant, aggegated_data)
                hybridlogger.ha_log(
                    self.logger, self.hass_api, "DEBUG", "Aggregated data processed."
                )

        # Finish datalogging process
        hybridlogger.ha_log(
            self.logger, self.hass_api, "DEBUG", "Timed data logging processed"
        )

        # Return the the time of the last report received
        return next_report_at

    def _process_pushed_event(self):
        # For localproxy client
        # Returns the date time of the last update
        next_report_at = None

        # Check for login, if needed. Return None on failure
        if not self._logon():
            # Skip this measurement, retry later
            return None

        # Caching of plant id's, if needed. Return None on failure
        if not self._fetch_plants():
            # Skip this measurement, retry later
            return None

        # Process data reports for plant event data
        if self.omnik_api_level == 2:
            # Wait for new data logging event
            data = self._listen_for_update()
            if data:
                plant = self._process_received_update(data)
                # Set last update time for plant
                self.plant_update[plant].last_update_time = datetime.fromtimestamp(
                    data["last_update"], timezone.utc
                )
                # Digitize data
                self._digitize(data)
                # Get specific dsmr data
                self._get_dsmr_data(plant, data)
                # Cache last update for aggregation
                self.plant_update[plant].data = data
                # export the data to the output plugins
                self._output_update(plant, data)
        else:
            return next_report_at

        # To aggregate over multiple inverters we need to trigger publishing when we an update for all our inverters
        # Unless the last data update of an inverter is longer then 2 timed cycles
        aggegated_data = {}
        for published_plant in self.plant_update:
            if self.plant_update[published_plant].pop_for_aggregate(self.cache):
                # Assemble aggegated data
                self._aggregate_data(
                    aggegated_data, self.plant_update[published_plant].data
                )
            else:
                # Not ready yet to aggregate, erase cache
                aggegated_data = {}
                break
        # only publish unpublished fresh data (<10 minutes old)
        if aggegated_data:
            # Get dsmr data for aggegated data
            self._total_energy_recalc(aggegated_data)
            self._get_dsmr_data("0", aggegated_data)
            # Output aggregated data to influx/mqtt if we have multiple plants
            if len(self.client.plant_id_list) > 1:
                self._output_update(plant, aggegated_data)
            self._output_update_aggregated_data("0", aggegated_data)
            #
            hybridlogger.ha_log(
                self.logger, self.hass_api, "DEBUG", "Aggregated data processed."
            )

        # Finish datalogging process
        hybridlogger.ha_log(
            self.logger, self.hass_api, "DEBUG", "Pushed logging processed"
        )

        # Return the the time of the latest report received
        return next_report_at

    def process(self):
        # Returns the date time of the last update (time behind) or a time in te future when postposing updates
        # Do a Sunshine check when a Timer is used
        if self.client.use_timer:
            return self._process_timed_event()
        else:
            # Return None when an error occurs
            return self._process_pushed_event()

    @staticmethod
    def __expand_path(path):
        """
        Expand relative path to absolute path.
        Args:
            path: file path
        Returns: absolute path to file
        """
        if os.path.isabs(path):
            return path
        else:
            return os.path.dirname(os.path.abspath(__file__)) + "/" + path



import json
import time
from omnik.ha_logger import hybridlogger

import urllib.parse
import requests
from omnik.plugin_output import Plugin
from decimal import Decimal
import threading


class pvoutput(Plugin):
    def __init__(self):
        super().__init__()
        self.name = "pvoutput"
        self.description = "Write output to PVOutput"
        # Enable support for aggregates
        self.process_aggregates = True
        # Make instance to run exclusively
        self.access = threading.Condition(threading.Lock())

    def _get_temperature(self, msg, data):
        if self.config.getboolean(
            "output.pvoutput", "use_temperature", fallback=True
        ) and self.config.getboolean(
            "output.pvoutput", "use_inverter_temperature", fallback=False
        ):
            if "inverter_temperature" in msg:
                data["v5"] = float(msg["inverter_temperature"])
            return
        if self.config.getboolean(
            "output.pvoutput", "use_temperature", fallback=False
        ) and not self.config.getboolean(
            "output.pvoutput", "use_inverter_temperature", fallback=False
        ):
            weather = self.get_weather()
            data["v5"] = weather["main"]["temp"]

    def _get_voltage(self, msg, data):
        voltage_field = self.config.get(
            "output.pvoutput", "publish_voltage", fallback=None
        )
        net_voltage_field = self.config.get(
            "output.pvoutput", "net_voltage_fallback", fallback=None
        )
        if voltage_field and voltage_field in msg:
            data["v6"] = float(msg[voltage_field])
        elif net_voltage_field and net_voltage_field in msg:
            data["v6"] = float(msg[net_voltage_field])

    def _check_requirements(self, msg):
        if not self.config.has_option("output.pvoutput", "api_key"):
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"[{__name__}] No api_key found in configuration",
            )
            return False

        if "sys_id" not in msg:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                f"[{__name__}] No sys_id found in dataset, set sys_id  "
                "in the output.pvoutput section",
            )
            return False
        return True

    def process(self, **args):
        """
        Send data to pvoutput
        """
        try:
            self.access.acquire()

            msg = args["msg"]
            if str(msg["sys_id"]) == "0":
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "DEBUG",
                    "ignoring publishing to pvoutput, no sys_id was set",
                )
                return
            reporttime = time.localtime(msg["last_update"])

            # self.logger.debug(json.dumps(msg, indent=2))
            if not self._check_requirements(msg):
                return

            headers = {
                "X-Pvoutput-Apikey": self.config.get("output.pvoutput", "api_key"),
                "X-Pvoutput-SystemId": str(msg["sys_id"]),
                "Content-type": "application/x-www-form-urlencoded",
                "Accept": "text/plain",
            }

            # see: https://pvoutput.org/help.html
            # see: https://pvoutput.org/help.html#api-addstatus
            # report cumulative values to enable power usage
            data = {
                "d": time.strftime("%Y%m%d", reporttime),
                "t": time.strftime("%H:%M", reporttime),
                "c1": 0,
            }
            # v1 = energy_generated (Wh) * 1000 ; this value is on a dayly basis
            # v2 = bruto power_generated (W)
            # v3 = energy_used (Wh) * 1000 ; this value is cumulative
            # v4 = bruto power_consumption (W)
            # c1 = 1 ; v3 is cumulative so the c1 flag is set
            # see note about Cumulative Enery at https://pvoutput.org/help.html#api-addstatus

            if "energy_used" in msg and "power_consumption" in msg:
                data.update(
                    {
                        "v3": f"{msg['energy_used'] * Decimal('1000')}",
                        "v4": f"{msg['power_consumption']}",
                        "c1": 1,
                    }
                )
                # v3 = energy_used (Wh) * 1000 ; this value is cumulative
                # v4 = bruto power_consumption (W)
                # c1 = 1 ; v3 is cumulative so the c1 flag is set
                # see note about Cumulative Enery at https://pvoutput.org/help.html#api-addstatus
            if "current_power" in msg:
                data.update(
                    {
                        "v1": f"{msg['today_energy'] * Decimal('1000')}",
                        "v2": f"{msg['current_power']}",
                    }
                )
                # v1 = energy_generated (Wh) * 1000 ; this value is on a dayly basis
                # v2 = bruto power_generated (W)
                # c1 = 0 ; v1 is not cumulative so the c1 flag not is set
                # see note about Cumulative Enery at https://pvoutput.org/help.html#api-addstatus

                # Publish (inverter) temperature if solar data is published and a temperature is available.
                # alternatively use the temperature from openweather
                self._get_temperature(msg, data)

            # Publish voltage (if available)
            self._get_voltage(msg, data)

            encoded = urllib.parse.urlencode(data)

            self.logger.debug(json.dumps(data, indent=2))

            r = requests.post(
                "http://pvoutput.org/service/r2/addstatus.jsp",
                data=encoded,
                headers=headers,
            )

            r.raise_for_status()

        except requests.exceptions.RequestException as err:
            hybridlogger.ha_log(
                self.logger, self.hass_api, "WARNING", f"Unhandled request error: {err}"
            )
        except Exception as e:
            hybridlogger.ha_log(
                self.logger, self.hass_api, "ERROR", f"Unexpected error: {e.args}"
            )
        finally:
            self.access.release()



#! /usr/bin/env python3

import os
import signal
import argparse
import configparser
import pathlib
import time
import logging
import yaml
from omnik.ha_logger import hybridlogger
from omnik import RepeatedJob
from omnik.datalogger import DataLogger

try:
    # AppDeamon is only needed when run used with AppDaemon
    import appdaemon.plugins.hass.hassapi as hass  # type: ignore
except ImportError:
    # AppDeamon is not needed when run in command mode uses dummy class
    class hass(object):
        Hass = object


logger = logging.getLogger(__name__)

# customize config parser with dict based lookup for AppDaemon and command line options
# has_option and get function have been adapted
ha_args = {}

global stopflag


# Generic signal handler
def signal_handler(signal, frame):
    global stopflag
    logging.debug("Signal {:0d} received. Setting stopflag.".format(signal))
    stopflag = True


class ha_ConfigParser(configparser.ConfigParser):
    def __init__(self, ha_args: dict = None, *args, **kwargs):
        if ha_args is None:
            self.ha_args = {}
        else:
            self.ha_args = ha_args
        super().__init__(*args, **kwargs)

    def has_option(self, section: str, option: str):
        if str(section).lower() == "default":
            if option in self.ha_args:
                return True
        else:
            if section in self.ha_args:
                if option in self.ha_args[section]:
                    return True
        return super().has_option(section, option)

    def get(self, section, option, fallback=None, **kwargs):
        if str(section).lower() == "default":
            if option in self.ha_args:
                return self.ha_args.get(option, fallback)
        else:
            if str(section) in self.ha_args:
                if option in self.ha_args[section]:
                    return self.ha_args[section].get(option, fallback)
        try:
            if fallback:
                retval = super().get(section, option, fallback=fallback, **kwargs)
            else:
                retval = super().get(section, option, fallback="", **kwargs)
        except Exception:
            retval = fallback
        return retval

    def getboolean(self, section, option, fallback=None, **kwargs):
        truelist = ["true", "1", "t", "y", "yes", "j", "ja"]
        valstr = str(self.get(section, option, fallback, **kwargs)).lower()
        return valstr in truelist

    def getlist(self, section, option, fallback=[], **kwargs):
        if str(section).lower() == "default":
            if option in self.ha_args:
                return self.ha_args.get(option, fallback)
        else:
            if str(section) in self.ha_args:
                if option in self.ha_args[section]:
                    return self.ha_args[section].get(option, fallback)
        try:
            retval = super().get(section, option, fallback=fallback, **kwargs)
        except Exception:
            retval = fallback
            pass
        return retval


# Initialization class for AppDaemon (homeassistant)
class HA_OmnikDataLogger(hass.Hass):
    def initialize(self, *args, **kwargs):
        hybridlogger.ha_log(logger, self, "INFO", "Starting Omnik datalogger...")
        hybridlogger.ha_log(
            logger,
            self,
            "DEBUG",
            f"Arguments from AppDaemon config (self.args): {self.args}",
        )
        if "config" in self.args:
            c = ha_ConfigParser(
                converters={"list": lambda x: [i.strip() for i in x.split(",")]},
                ha_args=self.args,
            )
            self.configfile = self.args["config"]
            try:
                c.read(self.configfile, encoding="utf-8")
                c.configfile = self.configfile
            except Exception as e:
                hybridlogger.ha_log(
                    logger,
                    self,
                    "ERROR",
                    f"Error parsing config.ini '{self.configfile}'. No valid configuration file. {e}.",
                )
                c.configfile = None
        else:
            c = ha_ConfigParser(ha_args=self.args)
            c.configfile = None
        set_data_config_path(c)

        self.interval = int(c.get("default", "interval", 360))
        self.datalogger = DataLogger(c, hass_api=self)

        # running repeatedly
        if self.datalogger.client.use_timer:
            hybridlogger.ha_log(
                logger, self, "INFO", f"Daemon interval {self.interval} seconds."
            )
        else:
            hybridlogger.ha_log(logger, self, "INFO", "Daemon starts listening.")
        self.rt = RepeatedJob(c, self.datalogger, hass_api=self)

    def terminate(self):
        hybridlogger.ha_log(logger, self, "INFO", "Stopping Omnikdatalogger...")
        self.rt.stop()
        self.datalogger.terminate()
        hybridlogger.ha_log(logger, self, "INFO", "Omnikdatalogger was stopped")


# Initialization from the commandline
def main(c: ha_ConfigParser, hass_api=None):
    global stopflag
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    # Enabled debugging if the flag is set
    if c.get("default", "debug", False):
        logger.setLevel(logging.DEBUG)

    set_data_config_path(c)
    datalogger = DataLogger(c, hass_api=hass_api)
    if c.has_option("default", "interval"):
        # running repeatedly
        if datalogger.client.use_timer:
            hybridlogger.ha_log(
                logger,
                hass_api,
                "INFO",
                f"Daemon interval {int(c.get('default', 'interval', 360))} seconds.",
            )
        else:
            hybridlogger.ha_log(logger, hass_api, "INFO", "Daemon starts listening.")
        rt = RepeatedJob(c, datalogger, hass_api)
        if hass_api:
            hass_api.rt = rt
        else:
            # loop is for commandline execution only
            try:
                while not stopflag:
                    time.sleep(0.5)
            except KeyboardInterrupt:
                pass
            hybridlogger.ha_log(logger, hass_api, "INFO", "Stopping Omnikdatalogger...")
            rt.stop()
            datalogger.terminate()
            hybridlogger.ha_log(logger, hass_api, "INFO", "Omnikdatalogger was stopped")

    else:
        # running only once
        datalogger.process()
        datalogger.terminate()


def set_data_config_path(config):
    config.data_config_file_path = (
        f"{pathlib.Path(__file__).parent.absolute()}/data_fields.json"
    )
    config.data_config_file_shared = (
        f"{pathlib.Path(__file__).parent.parent.absolute()}"
        "/share/omnikdatalogger/data_fields.json"
    )
    config.data_config_file_cwd = f"{os.getcwd()}/data_fields.json"


def get_yaml_settings(args):
    with open(args.settings, "r") as stream:
        try:
            settings = yaml.safe_load(stream)
            if args.section in settings:
                index = args.section
            else:
                for key in settings:
                    index = key
                    break
            logging.info(
                "Using section '{1}' from config file '{0}'".format(
                    args.settings, index
                )
            )
            return settings[index]
        except yaml.YAMLError as exc:
            logging.error(
                "YAML config file '{0}' could not be parsed. Error: {1}".format(
                    args.settings, exc
                )
            )
            os.sys.exit(1)


def setup_config_parser(args, settings):
    if not settings:
        logging.error("Config files '{0}' has no valid index".format(args.settings))
        os.sys.exit(1)
    elif "config" in settings:
        try:
            configfile = settings["config"]
            c = ha_ConfigParser(
                converters={"list": lambda x: [i.strip() for i in x.split(",")]},
                ha_args=settings,
            )
            c.read(configfile, encoding="utf-8")
            c.configfile = configfile
            logging.info(
                "Using config file '{0}' from the 'config' key of YAML file {1}".format(
                    configfile, args.settings
                )
            )
        except Exception as e:
            logging.error(
                f"Error parsing config.ini '{configfile}'. No valid configuration file. {e}."
            )
            os.sys.exit(1)
    else:
        try:
            c = ha_ConfigParser(ha_args=settings)
            c.configfile = None
        except Exception as e:
            logging.error("Config parser problem '{0}'".format(e))
            os.sys.exit(1)
    return c


def get_config_from_files(args):
    c = None
    if args.debug:
        ha_args["debug"] = args.debug
    if args.interval:
        ha_args["interval"] = args.interval
    if args.data_config:
        ha_args["data_config"] = args.data_config
    if args.persistant_cache_file:
        ha_args["persistant_cache_file"] = args.persistant_cache_file
    if os.path.isfile(args.settings):
        if os.path.isfile(args.config):
            logging.warning(
                "Multiple config files found at '{0}' or '{1}', "
                "using '{0}', ignoring config from command line option '{1}'.'".format(
                    args.settings, args.config
                )
            )

        settings = get_yaml_settings(args)
        # commandline override
        settings.update(ha_args)
        c = setup_config_parser(args, settings)
    if os.path.isfile(args.config) and not os.path.isfile(args.settings):
        ha_args["config"] = args.config
        c = ha_ConfigParser(
            converters={"list": lambda x: [i.strip() for i in x.split(",")]},
            ha_args=ha_args,
        )
        c.read([args.config], encoding="utf-8")
        c.configfile = args.config
    if not c:
        logging.error("No valid configuration found. Exiting!")
        os.sys.exit(1)
    return c


if __name__ == "__main__":
    stopflag = False
    home = os.path.expanduser("~")
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--settings",
        default=os.path.join(home, ".omnik/config.yaml"),
        help="Path to .yaml configuration file",
        metavar="FILE",
    )
    parser.add_argument(
        "--section",
        default=None,
        help="Section to .yaml configuration file to use. Defaults to the first section found.",
    )
    parser.add_argument(
        "--config",
        default=os.path.join(home, ".omnik/config.ini"),
        help="Path to configuration file (ini) (DECREPATED!)",
        metavar="FILE",
    )
    parser.add_argument(
        "--data_config",
        help="Path to data_fields.json configuration file",
        metavar="FILE",
    )
    parser.add_argument(
        "--persistant_cache_file",
        help="Path to writable cache json file to store last power en total energy",
        metavar="FILE",
    )
    parser.add_argument("--interval", type=int, help="Execute every n seconds")

    parser.add_argument("-d", "--debug", action="store_true", help="Debug mode")

    args = parser.parse_args()

    if not os.path.isfile(args.settings) and not os.path.isfile(args.config):
        logging.critical(
            "Unable to find config at '{0}' or {1}".format(args.settings, args.config)
        )
        parser.print_help()
        os.sys.exit(1)
    elif os.path.isfile(args.config):
        logging.warning(
            "The use of a classig config.ini is decrepated! "
            "Use the --settings option. See documentation."
        )

    main(get_config_from_files(args), hass_api=None)



from omnik.ha_logger import hybridlogger
from omnik.plugin_localproxy import LocalProxyPlugin
import binascii


class HASSAPI(LocalProxyPlugin):

    """
    This plugin enables you to subscribe to the logger requests using the AppDaemon HASSAPI integration
    Use scripts/proxy/omnikloggerproxy.py to capture the inverter events and publish the data to MQTT (Home Assistant)

    This class makes use of the following localproxy client attributes
    self.client.msg
    self.client.msgevent
    self.client.semaphore
    self.client.inverters
    self.client.plant_id_list
    """

    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(
            self.logger, self.hass_api, "INFO", "localproxy client plugin: HASSAPI"
        )
        self.hass_handle = {}
        if not self.hass_api:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                "No HassAPI detected. Use AppDaemon with Home Assistent for this plugin",
            )
            return
        # Deprecated property (use only if no plant specific configuration is available)
        self.legacy_logger_entity = self.config.get(
            "client.localproxy.hassapi", "logger_entity", "binary_sensor.datalogger"
        )
        self.logger_entity = []
        for plant in self.client.plant_id_list:
            entity = self.config.get(f"plant.{plant}", "logger_entity", None)
            if entity:
                self.logger_entity.append(entity)
        if self.legacy_logger_entity not in self.logger_entity:
            self.logger_entity.append(self.legacy_logger_entity)

    def terminate(self):
        if self.hass_handle:
            for handle in self.hass_handle:
                self.hass_api.cancel_listen_state(self.hass_handle[handle])

    def listen(self):
        if self.hass_api:
            for logger_entity in self.logger_entity:
                self.hass_handle[logger_entity] = self.hass_api.listen_state(
                    self._run, logger_entity, attribute="data"
                )
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "INFO",
                    f"HASSapi listening to. '{logger_entity}', attribute: 'data'",
                )

    def _run(self, entity, attribute, old, new, kwargs):
        # HASSAPI callback handler
        hybridlogger.ha_log(
            self.logger,
            self.hass_api,
            "INFO",
            f"HASSapi state change for {self.hass_api.get_state(entity, 'inverter', 'n/a')} "
            f"at {self.hass_api.get_state(entity, 'last_update', 'n/a')}",
        )
        # Try to parse payload as json
        try:
            data = binascii.a2b_base64(new)
            if len(data) == 128:
                # Fill data structure
                self.client.semaphore.acquire()
                self.client.msg["data"] = data
                self.client.msg["isSet"] = True
                self.client.msg["plugin"] = __name__
                self.client.semaphore.release()
                # Trigger processing the message
                self.client.msgevent.set()
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"HASSapi: invalid data received: {new}. Error {e}",
            )



from datetime import datetime, timezone, time
from decimal import Decimal


class Plant:
    def __init__(self, plant_id=None, last_update_time=None, *args, **kwargs):
        self._plant_id = plant_id
        self._last_update_time = (
            last_update_time if last_update_time else datetime.now(timezone.utc)
        )
        self._updated = bool(last_update_time)
        self._data = None

    def plant_id(self):
        return self._plant_id

    def pop_for_aggregate(self, cache):
        data_age = (datetime.now(timezone.utc) - self._last_update_time).seconds
        if data_age > 360:
            # data is too old discard, use cache
            self._updated = False
            # return True to ensure publising
            if not self._data:
                self._data = {}
            try:
                self._data["total_energy"] = cache[
                    f"{self._plant_id}.last_total_energy"
                ]
                self._data["today_energy"] = cache[
                    f"{self._plant_id}.last_today_energy"
                ]
                self._data["current_power"] = Decimal("0.0")
                self._data["last_update"] = time.time()
            except:
                return False
            return True
        if self._updated:
            self._updated = False
            return True
        else:
            # Data still valid, wait until all inverters showed their data
            return False

    @property
    def last_update_time(self):
        return self._last_update_time

    @last_update_time.setter
    def last_update_time(self, last_update_time):
        self._last_update_time = last_update_time
        self._updated = True

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self._data = data
        self._updated = True
